import * as fs from "fs"
import * as vm from "vm"
import * as globalConfig from "./globalConfig"
import { Watcher } from "./watcher"
import * as pathModule from "path"
import * as child_process from "child_process"
import * as wrench from "wrench"
import { Queue } from "./queue"
import { Rule } from "./rule"
import { ChangeMap } from "./changeMap"
import * as DateString from "./dateString"
import * as commander from "commander"


let program = commander
    .option("-f,--file <filename>", "specifail the watchfile default is Watchfile")
    .option("-c,--create-default", "create and default rules save as ./Watchfile")
    .option("-v,--version", "print version")
    .option("-s,--start-compile", "compile all matched file at start")
    .option("-q,--quit", "combined with -s, quit program after start compile")
    .option("--hash-check", "don't check file content hash change")
    .option("--verbose", "be verbose")
    .option("--shell <shell-path>", "exec shell path")
    .version("0.0.5")
    .parse(process.argv);

let defaultWatchFile = `//{basename} /css/style.less => style.less
//{fullpath} /css/style.less => /css/style.less (unchanged)
//{filename} /css/style.less => style
//{extname}  /css/style.less => .less
//{directory} /css/style.less => /css/
exports.watchList = [
    // [testFunctionOrRegExp,commandToRun]
    // [RegExp|(path:string)=>boolean,string]
    [/^.*\.coffee$/,'coffee -c {fullpath}'],
    [/^.*\.less$/,'lessc {fullpath} > {directory}{filename}.css'],
];

exports.serviceList = [
    //commandToRunOnceAtStart
    "echo watchcompile start",
]
`

if (program.createDefault) {
    console.log("create default Watchfile at ./Watchfile")
    if (fs.existsSync("./Watchfile")) {
        console.error("./Watchfile exists, don't overwrite it.")
        process.exit(1)
    }
    fs.writeFileSync("./Watchfile", defaultWatchFile)
    console.log("Done creating default watchfile")
}


// avoid Warnning -> possible EventEmitter memory leak detected. 11 listeners added.Use emitter.setMaxListeners() to increase limit.
process.stdout.setMaxListeners(2000)
process.stderr.setMaxListeners(2000)
const ignoreHidden = !program.all;
const watchFile = program.file || "./Watchfile";
const noHashCheck = !program.noHashCheck && true || false
let list: any[] = null
let serviceList: any[] = null
let watchFolders: any[] = null
let blacklist: (string | RegExp)[] = []
try {
    const context = vm.createContext({ exports: {}, console })
    const WatchfileCode = fs.readFileSync(watchFile, "utf8")
    vm.runInContext(WatchfileCode, context, "watchFile")
    list = context.exports.watchList || []
    serviceList = context.exports.serviceList || []
    watchFolders = context.exports.watches && context.exports.watches.map(item => pathModule.join(pathModule.dirname(watchFile), item)) || [pathModule.dirname(pathModule.resolve(watchFile))]
    blacklist = context.exports.blacklist || []
} catch (e) {
    console.error("invalid watchfile '%s'", watchFile)
    process.exit(1)
}

console.log("Watching:", watchFolders)
for (let service of serviceList) {
    const cp = child_process.exec(service)
    cp.stdout.pipe(process.stdout)
    cp.stderr.pipe(process.stderr)
}

if (program.shell) {
    // @ts-ignore
    globalConfig.shell = program.shell
    console.log("Using shell", globalConfig.shell)
}
let rules = []
let queue = new Queue(program.verbose)
let changeMap = new ChangeMap()
for (let config of list) {
    rules.push(new Rule(config))
}

// Calculate watch dir
if (typeof watchFolders == "string") {
    watchFolders = [watchFolders]
}
if (!watchFolders.forEach) {
    throw new Error("Invalid watchfolder")
}
let watchFileFolder = pathModule.dirname(watchFile)
//watchFolders = watchFolders.map folder => pathModule.join(watchFileFolder, folder)
function isPathBlacklisted(path: string) {
    for (let condition of blacklist) {
        if (typeof condition == "string") {
            if (path.indexOf(condition) >= 0) return true
            continue
        }
        else if (condition instanceof RegExp) {
            if (condition.test(path)) {
                return true
            }
            continue
        }
    }
    return false
}

if (program.startCompile) {
    for (let folder of watchFolders) {
        let files = wrench.readdirSyncRecursive(pathModule.join(watchFileFolder), folder)
        for (let path of files) {
            path = pathModule.join(folder, path)
            if (isPathBlacklisted(path)) {
                continue
            }
            for (let rule of rules) {
                if (rule.test(path)) {
                    let task = rule.taskFromPath(path)
                    console.log(`${DateString.genReadableDateString()} create ${task.toString()} by ${path}: inital compile`)
                    queue.add(task)
                }
            }
        }
    }
}
let watcher = new Watcher(watchFolders, isPathBlacklisted)

watcher.events.on("change", (path) => {
    if (!noHashCheck && !changeMap.checkAndUpdate(path)) {
        if (program.verbose) {
            console.log("#{path} file changed but content hash doesn't, skip it.")
        }
        return
    }
    for (let rule of rules) {
        if (rule.test(path)) {
            let task = rule.taskFromPath(path)
            console.log(`${DateString.genReadableDateString()} create ${task.toString()}}: modification`)
            queue.add(task)
        }
    }
})

queue.events.on("empty", () => {
    if (program.startCompile && program.quit) {
        console.log("start compile done and quit by -q")
        process.exit(0)
    }
})
if (!program.quit) {
    console.log("start watching")
}
